@@ -9,43 +9,32 @@
  * This file is part of the Math::Prime::Util Perl module.  It is not too hard
  * to build this as a standalone program (see the README file).
  *
- * This is pretty good for numbers less than 500 digits.  It is many orders
+ * This is pretty good for numbers less than 800 digits.  It is many orders
  * of magnitude faster than the contemporary GMP-ECPP.  It is far, far slower
- * than PRIMO.  I suspect it is slower than the 10-20 year old work of
- * François Morain.
+ * than PRIMO for numbers over 300 digits (it's a bit faster below that).  It
+ * is missing optimizations that were in François Morain's 10-20 year old work.
  *
  * A set of fixed discriminants are used, rather than calculating them as
  * needed.  Having a way to calculate values as needed would be a big help.
  * In the interests of space for the MPU package, I've chosen ~500 values which
  * compile into about 35k of data.  This is about 1/5 of the entire code size
- * for the MPU package.  The github repository includes a alternate set of 2650
- * discriminants that compile to 1.2MB.  This would be helpful if proving
- * 300+ digit numbers is a regular occurance.
+ * for the MPU package.  The github repository includes a alternate set of 3182
+ * discriminants that compile to 2.3MB.  This is recommended if proving 300+
+ * digit numbers is a regular occurance.
  *
  * This version uses the FAS "factor all strategy", meaning it first constructs
  * the entire factor chain, with backtracking if necessary, then will do the
  * elliptic curve proof as it recurses back.
  *
- * Another across-the-board performance improvement could be had by linking in
- * the GMP-ECM factoring package, which is much faster than my N-1 and ECM.
- * I have not yet measured what this would do.
- *
  * If your goal is primality proofs for very large numbers, use Primo.  It's
  * free, it is really fast, it is widely used, it can process batch results,
- * and it makes independently verifiable certificates.  If you want proofs
- * for "small" numbers (~300 digits) then this software should work well.
- * It also allows tinkering with the program as desired.
+ * and it makes independently verifiable certificates.  This software also
+ * makes certificates in a described format, works well under 800 or so digits,
+ * and is open source.
  *
- * Benchmarks.  These are using the small (500 values) class polynomial set.
- * Times are on a single core of a 4.2GHz i7-3930K (a fast machine), from
- * the command line using Perl, averaged across multiple runs.
- *
- *    10**49+9        0.007s
- *    10**100+267     0.04s
- *    2**511+111      0.10s
- *    2**1023+1155    2.1s
- *    2**2067+131    72s
- *    10**1000+453  9 minutes
+ * There are other proof methods, such as the open source APR-CL code from
+ * WraithX.  For numbers over 1000 digits, the APR-CL implementation will
+ * likely be faster than this, though does not produce a certificate.
  *
  * Thanks to H. Cohen, R. Crandall & C. Pomerance, and H. Riesel for their
  * text books.  Thanks to the authors of open source software who allow me
@@ -53,8 +42,9 @@
  * Thanks to Schoof, Goldwasser, Kilian, Atkin, Morain, Lenstra, etc. for all
  * the math and publications.  Thanks to Gauss, Euler, et al.
  *
- * The ECM code in ecm.c leverages early GMP-ECM work by Phil Zimmerman, as
- * well as all the articles of Montgomery, Bosma, Lentra, Cohen, and others.  
+ * The ECM code in ecm.c was heavily influenced by early GMP-ECM work by Phil
+ * Zimmerman, as well as all the articles of Montgomery, Bosma, Lentra,
+ * Cohen, and others.
  */
 
 #include <stdio.h>
@@ -78,7 +68,42 @@
  #include <ecm.h>
 #endif
 
-static int check_for_factor2(mpz_t f, mpz_t inputn, mpz_t fmin, mpz_t n, int stage, mpz_t* sfacs, int* nsfacs)
+#ifdef USE_APRCL
+ #include "mpz_aprcl.h"
+ #include "mpz_aprcl.c"
+#endif
+
+/*********** big primorials and lcm for divisibility tests  **********/
+static int _gcdinit = 0;
+static mpz_t _gcd_small;
+static mpz_t _gcd_large;
+static mpz_t _lcm_small;
+
+void init_ecpp_gcds(void) {
+  if (_gcdinit == 0) {
+    mpz_init(_gcd_small);
+    mpz_init(_gcd_large);
+    _GMP_pn_primorial(_gcd_small,  3000);
+    _GMP_pn_primorial(_gcd_large, 20000);
+    mpz_divexact(_gcd_large, _gcd_large, _gcd_small);
+    mpz_divexact_ui(_gcd_small, _gcd_small, 2*3*5);
+    mpz_init(_lcm_small);
+    _GMP_lcm_of_consecutive_integers(300, _lcm_small);
+    _gcdinit = 1;
+  }
+}
+
+void destroy_ecpp_gcds(void) {
+  if (!_gcdinit) return;
+  mpz_clear(_gcd_small);
+  mpz_clear(_gcd_large);
+  mpz_clear(_lcm_small);
+  _gcdinit = 0;
+}
+
+
+
+static int check_for_factor2(mpz_t f, mpz_t inputn, mpz_t fmin, mpz_t n, int stage, mpz_t* sfacs, int* nsfacs, int degree)
 {
   int success, sfaci;
   UV B1;
@@ -88,8 +113,9 @@ static int check_for_factor2(mpz_t f, mpz_t inputn, mpz_t fmin, mpz_t n, int sta
 
   if (mpz_cmp(n, fmin) <= 0) return 0;
 
+#if 0
   {
-    /* Alternate trial division: */
+    /* Straightforward trial division up to 3000. */
     PRIME_ITERATOR(iter);
     UV tf;
     UV const trial_limit = 3000;
@@ -100,20 +126,53 @@ static int check_for_factor2(mpz_t f, mpz_t inputn, mpz_t fmin, mpz_t n, int sta
     }
     prime_iterator_destroy(&iter);
   }
+#else
+  /* Utilize GMP's fast gcd algorithms.  Trial to 224737 with two gcds. */
+  mpz_tdiv_q_2exp(n, n, mpz_scan1(n, 0));
+  while (mpz_divisible_ui_p(n, 3))  mpz_divexact_ui(n, n, 3);
+  while (mpz_divisible_ui_p(n, 5))  mpz_divexact_ui(n, n, 5);
+  if (mpz_cmp(n, fmin) <= 0) return 0;
+  mpz_gcd(f, n, _gcd_small);
+  while (mpz_cmp_ui(f, 1) > 0) {
+    mpz_divexact(n, n, f);
+    mpz_gcd(f, n, _gcd_small);
+  }
+  if (mpz_cmp(n, fmin) <= 0) return 0;
+  mpz_gcd(f, n, _gcd_large);
+  while (mpz_cmp_ui(f, 1) > 0) {
+    mpz_divexact(n, n, f);
+    mpz_gcd(f, n, _gcd_large);
+  }
+  /* Quick stage 1 n-1 using a single big powm + gcd. */
+  if (stage == 0) {
+    if (mpz_cmp(n, fmin) <= 0) return 0;
+    mpz_set_ui(f, 2);
+    mpz_powm(f, f, _lcm_small, n);
+    mpz_sub_ui(f, f, 1);
+    mpz_gcd(f, f, n);
+    if (mpz_cmp_ui(f, 1) != 0 && mpz_cmp(f, n) != 0) {
+      mpz_divexact(n, n, f);
+      if (mpz_cmp(f, n) > 0)
+        mpz_set(n, f);
+    }
+  }
+#endif
 
   sfaci = 0;
   success = 1;
   while (success) {
+    UV nsize = mpz_sizeinbase(n, 2);
 
     if (mpz_cmp(n, fmin) <= 0) return 0;
     if (_GMP_is_prob_prime(n)) { mpz_set(f, n); return (mpz_cmp(f, fmin) > 0); }
 
     success = 0;
-    B1 = 300 + 3 * mpz_sizeinbase(n, 2);
+    B1 = 300 + 3 * nsize;
+    if (degree <= 2) B1 += nsize; /* D1 & D2 are cheap to prove.  Encourage. */
+    if (degree <= 0) B1 += nsize; /* N-1 and N+1 are really cheap. */
+    if (degree > 20 && stage <= 1) B1 -= nsize;   /* Less time on big polys. */
+    if (degree > 40) B1 -= nsize/2;               /* Less time on big polys. */
     if (stage >= 1) {
-      /* Push harder for big numbers.  (1) to avoid backtracking too much, and
-       * (2) keep poly degree down to save time in root finding later. */
-      /* Alternate:  UV B1 = (mpz_sizeinbase(n, 2) > 1200) ? 6000 : 3000; */
 #ifdef USE_LIBECM
       /* TODO: Tune stage 1 (PM1?) */
       /* TODO: LIBECM in other stages */
@@ -128,7 +187,9 @@ static int check_for_factor2(mpz_t f, mpz_t inputn, mpz_t fmin, mpz_t n, int sta
         if (mpz_cmp(f, n) == 0)  success = 0;
       }
 #else
-      if (!success) success = _GMP_pminus1_factor(n, f, B1, 10*B1);
+      if (!success) success = _GMP_pminus1_factor(n, f, B1, 6*B1);
+      if (!success) success = _GMP_pplus1_factor(n, f, 0, B1/8, B1/8);
+      if (!success && nsize < 500) success = _GMP_pbrent_factor(n, f, nsize, 1024-nsize);
 #endif
     }
     /* Try any factors found in previous stage 2+ calls */
@@ -141,17 +202,23 @@ static int check_for_factor2(mpz_t f, mpz_t inputn, mpz_t fmin, mpz_t n, int sta
     }
     if (stage > 1 && !success) {
       if (stage == 2) {
-        if (!success) success = _GMP_pminus1_factor(n, f, 5*B1, 5*20*B1);
-        if (!success) success = _GMP_ecm_factor_projective(n, f, 250, 4);
+        if (!success) success = _GMP_pbrent_factor(n, f, nsize-1, 8192);
+        if (!success) success = _GMP_pminus1_factor(n, f, 6*B1, 60*B1);
+        /* p+1 with different initial point and searching farther */
+        if (!success) success = _GMP_pplus1_factor(n, f, 1, B1/2, B1/2);
+        if (!success) success = _GMP_ecm_factor_projective(n, f, 250, 3);
       } else if (stage == 3) {
-        if (!success) success = _GMP_pminus1_factor(n, f, 25*B1, 25*20*B1);
-        if (!success) success = _GMP_ecm_factor_projective(n, f, 500, 4);
+        if (!success) success = _GMP_pbrent_factor(n, f, nsize+1, 16384);
+        if (!success) success = _GMP_pminus1_factor(n, f, 60*B1, 600*B1);
+        /* p+1 with a third initial point and searching farther */
+        if (!success) success = _GMP_pplus1_factor(n, f, 2, 1*B1, 1*B1);
+        if (!success) success = _GMP_ecm_factor_projective(n, f, B1/4, 4);
       } else if (stage == 4) {
-        if (!success) success = _GMP_pminus1_factor(n, f, 200*B1, 200*20*B1);
-        if (!success) success = _GMP_ecm_factor_projective(n, f, 1000, 10);
+        if (!success) success = _GMP_pminus1_factor(n, f, 300*B1, 300*20*B1);
+        if (!success) success = _GMP_ecm_factor_projective(n, f, B1/2, 4);
       } else if (stage >= 5) {
-        UV B = 8000 * (stage-4) * (stage-4) * (stage-4);
-        if (!success) success = _GMP_ecm_factor_projective(n, f, B, 5+stage);
+        UV B = B1 * (stage-4) * (stage-4) * (stage-4);
+        if (!success) success = _GMP_ecm_factor_projective(n, f, B, 8+stage);
       }
     }
     if (success) {
@@ -269,29 +336,44 @@ static void weber_root_to_hilbert_root(mpz_t r, mpz_t N, long D)
   mpz_init(A);  mpz_init(t);
 
   switch (D % 8) {
-    case 1:  mpz_powm_ui(t, r, 12, N);
+    case 1:  if ((D % 3) != 0)  mpz_powm_ui(t, r, 12, N);
+             else               mpz_powm_ui(t, r,  4, N);
              mpz_mul_ui(A, t, 64);
              mpz_sub_ui(t, A, 16);
              break;
     case 2:
-    case 6:  mpz_powm_ui(t, r, 12, N);
+    case 6:  if ((D % 3) != 0)  mpz_powm_ui(t, r, 12, N);
+             else               mpz_powm_ui(t, r,  4, N);
              mpz_mul_ui(A, t, 64);
              mpz_add_ui(t, A, 16);
              break;
-    case 5:  mpz_powm_ui(t, r, 6, N);
+    case 5:  if ((D % 3) != 0)  mpz_powm_ui(t, r, 6, N);
+             else               mpz_powm_ui(t, r, 2, N);
              mpz_mul_ui(A, t, 64);
              mpz_sub_ui(t, A, 16);
              break;
     case 7:  if (!mpz_invert(t, r, N)) mpz_set_ui(t, 0);
-             mpz_powm_ui(A, t, 24, N);
+             if ((D % 3) != 0)  mpz_powm_ui(A, t, 24, N);
+             else               mpz_powm_ui(A, t,  8, N);
+             mpz_sub_ui(t, A, 16);
+             break;
+    /* Results in degree 3x Hilbert, so typically not used */
+    case 3:  if (!mpz_invert(t, r, N)) mpz_set_ui(t, 0);
+             if ((D % 3) != 0) {
+               mpz_powm_ui(t, t, 24, N);
+               mpz_mul_2exp(A, t, 12);
+             } else {
+               mpz_powm_ui(t, t, 8, N);
+               mpz_mul_2exp(A, t, 4);
+             }
              mpz_sub_ui(t, A, 16);
              break;
     default: break;
   }
+  /* r = t^3 / A */
   mpz_powm_ui(t, t, 3, N);
-  if (!mpz_invert(A, A, N)) mpz_set_ui(A, 0);
-  mpz_mulmod(r, A, t, N, r);
-
+  if ( ! mpz_divmod(r, t, A, N, r) )
+    mpz_set_ui(r, 0);
   mpz_clear(A);  mpz_clear(t);
 }
 
@@ -560,12 +642,14 @@ static void choose_m(mpz_t* mlist, long D, mpz_t u, mpz_t v, mpz_t N,
  */
 
 /* Recursive routine to prove via ECPP */
-static int ecpp_down(int i, mpz_t Ni, int facstage, UV* dlist, mpz_t* sfacs, int* nsfacs, char** prooftextptr)
+static int ecpp_down(int i, mpz_t Ni, int facstage, IV* dlist, mpz_t* sfacs, int* nsfacs, char** prooftextptr)
 {
-  mpz_t a, b, u, v, m, q, minfactor, mD, t, t2;
+  mpz_t a, b, u, v, m, q, minfactor, sqrtn, mD, t, t2;
   mpz_t mlist[6];
+  UV nm1a;
+  IV np1lp, np1lq;
   struct ec_affine_point P;
-  int k, dnum, nidigits, facresult, curveresult, downresult, stage;
+  int k, dnum, nidigits, facresult, curveresult, downresult, stage, D;
   int verbose = get_verbose_level();
 
   nidigits = mpz_sizeinbase(Ni, 10);
@@ -579,45 +663,6 @@ static int ecpp_down(int i, mpz_t Ni, int facstage, UV* dlist, mpz_t* sfacs, int
   }
   downresult = 0;
 
-#ifdef USE_NM1
-    /* Try a simple n-1 test if the number is small enough to give it a decent
-     * chance of success.
-     *
-     * For not-big numbers, effort 1 is nearly free, 2 is very fast, 3 is ok.
-     * For effort 2, it is about 50% at 128 bits, 25% at 160, 10% at 190. */
-    {
-      UV log2ni = mpz_sizeinbase(Ni, 2);
-      int effort = (log2ni <= 128) ? 2
-                 : (log2ni <= 256) ? 1
-                 :                   0;
-      if (effort > 0) {
-        char* nm1_proof = 0;
-        /* -1 = composite, 0 = dunno, 1 = proved prime */
-        k = _GMP_primality_bls_nm1(Ni, effort, &nm1_proof);
-        if (k == 2) {
-          if (prooftextptr != 0 && nm1_proof != 0) {
-            char *proofstr;
-            int myprooflen = strlen(nm1_proof);
-            int curprooflen = (*prooftextptr == 0) ? 0 : strlen(*prooftextptr);
-            Newz(0, proofstr, myprooflen + curprooflen + 1, char);
-            (void) strcat(proofstr, nm1_proof);
-            if (*prooftextptr) {
-              strcat(proofstr + myprooflen, *prooftextptr);
-              Safefree(*prooftextptr);
-            }
-            *prooftextptr = proofstr;
-          }
-          if (verbose)
-            printf("%*sN[%d] (%d dig)  PRIME (n-1)\n", i, "", i, nidigits);
-        }
-        if (nm1_proof != 0)
-          Safefree(nm1_proof);
-        if (k != 1)
-          return k;
-      }
-    }
-#endif
-
   if (verbose) {
     printf("%*sN[%d] (%d dig)", i, "", i, nidigits);
     if (facstage > 1) printf(" FS %d", facstage);
@@ -627,7 +672,7 @@ static int ecpp_down(int i, mpz_t Ni, int facstage, UV* dlist, mpz_t* sfacs, int
   mpz_init(a);  mpz_init(b);
   mpz_init(u);  mpz_init(v);
   mpz_init(m);  mpz_init(q);
-  mpz_init(mD); mpz_init(minfactor);
+  mpz_init(mD); mpz_init(minfactor);  mpz_init(sqrtn);
   mpz_init(t);  mpz_init(t2);
   mpz_init(P.x);mpz_init(P.y);
   for (k = 0; k < 6; k++)
@@ -638,27 +683,100 @@ static int ecpp_down(int i, mpz_t Ni, int facstage, UV* dlist, mpz_t* sfacs, int
   mpz_root(minfactor, Ni, 4);
   mpz_add_ui(minfactor, minfactor, 1);
   mpz_mul(minfactor, minfactor, minfactor);
+  mpz_sqrt(sqrtn, Ni);
 
   for (stage = (i == 0) ? facstage : 1; stage <= facstage; stage++) {
-    for (dnum = 0; dlist[dnum] > 0; dnum++) {
+    int next_stage = (stage > 1) ? stage : 1;
+    for (dnum = 0; dlist[dnum] != 0; dnum++) {
       int poly_type;  /* just for debugging/verbose */
       int poly_degree;
-      long D = -dlist[dnum];
-      if (D == -1) continue;  /* Marked for skip */
+      D = -dlist[dnum];
+      if (D > 1) continue;  /* Marked for skip */
+
+      if (D == 1) {   /* n-1 test */
+        mpz_sub_ui(m, Ni, 1);          /* m = N-1 */
+        mpz_sub_ui(t2, sqrtn, 1);
+        mpz_tdiv_q_2exp(t2, t2, 1);    /* t2 = minfactor */
+
+        facresult = check_for_factor2(q, m, t2, t, stage, sfacs, nsfacs, 0);
+        if (facresult <= 0)
+          continue;
+        if (verbose)
+          { printf(" n-1\n"); fflush(stdout); }
+        downresult = ecpp_down(i+1, q, next_stage, dlist, sfacs, nsfacs, prooftextptr);
+        if (downresult == 0)     /* composite */
+          goto end_down;
+        if (downresult == 1) {   /* nothing found at this stage */
+          if (verbose) {
+            printf("%*sN[%d] (%d dig)", i, "", i, nidigits);
+            if (facstage > 1) printf(" FS %d", facstage);
+            fflush(stdout);
+          }
+          continue;
+        }
+        if (verbose)
+          { printf("%*sN[%d] (%d dig) n-1", i, "", i, nidigits); fflush(stdout); }
+        curveresult = _GMP_primality_bls_3(Ni, q, &nm1a);
+        if (verbose) { printf("  %d\n", curveresult); fflush(stdout); }
+        if ( ! curveresult ) {
+          /* This ought not happen */
+          dlist[dnum] = -2; /* skip this D value from now on */
+          if (verbose) gmp_printf("\n  Could not prove n-1 with N = %Zd\n", D, Ni);
+          downresult = 0;
+          continue;
+        }
+        goto end_down;
+      }
+      if (D == -1) {  /* n+1 test */
+        mpz_add_ui(m, Ni, 1);          /* m = N+1 */
+        mpz_add_ui(t2, sqrtn, 1);
+        mpz_tdiv_q_2exp(t2, t2, 1);    /* t2 = minfactor */
+
+        facresult = check_for_factor2(q, m, t2, t, stage, sfacs, nsfacs, 0);
+        if (facresult <= 0)
+          continue;
+        if (verbose)
+          { printf(" n+1\n"); fflush(stdout); }
+        downresult = ecpp_down(i+1, q, next_stage, dlist, sfacs, nsfacs, prooftextptr);
+        if (downresult == 0)     /* composite */
+          goto end_down;
+        if (downresult == 1) {   /* nothing found at this stage */
+          if (verbose) {
+            printf("%*sN[%d] (%d dig)", i, "", i, nidigits);
+            if (facstage > 1) printf(" FS %d", facstage);
+            fflush(stdout);
+          }
+          continue;
+        }
+        if (verbose)
+          { printf("%*sN[%d] (%d dig) n-1", i, "", i, nidigits); fflush(stdout); }
+        curveresult = _GMP_primality_bls_15(Ni, q, &np1lp, &np1lq);
+        if (verbose) { printf("  %d\n", curveresult); fflush(stdout); }
+        if ( ! curveresult ) {
+          /* This ought not happen */
+          dlist[dnum] = -2; /* skip this D value from now on */
+          if (verbose) gmp_printf("\n  Could not prove n-1 with N = %Zd\n", D, Ni);
+          downresult = 0;
+          continue;
+        }
+        goto end_down;
+      }
+
       if ( (-D % 4) != 3 && (-D % 16) != 4 && (-D % 16) != 8 )
-        croak("Invalid discriminant '%ld' in list\n", D);
+        croak("Invalid discriminant '%d' in list\n", D);
       /* D must also be squarefree in odd divisors, but assume it. */
       /* Make sure we can get a class polynomial for this D. */
       poly_degree = poly_class_poly(D, NULL, &poly_type);
       if (poly_degree == 0)  continue;
       /* We'll save time in the long run by not looking at big polys once
        * we've found a good path from the start.  TODO: Needs more tuning. */
+      if (stage == 0 && i >= 0 && poly_degree > 2) break;
       if (facstage == 1) {
-        if (i >  2 && poly_degree > 24)  break;
-        if (i >  3 && poly_degree > 16)  break;
-        if (i >  4 && nidigits < 800 && poly_degree > 12)  break;
-        if (i >  8 && nidigits < 700 && poly_degree > 10)  break;
-        if (i > 16 && nidigits < 600 && poly_degree >  8)  break;
+        if (i >  2 && nidigits < 1100 && poly_degree > 24)  break;
+        if (i >  3 && nidigits <  950 && poly_degree > 15)  break;
+        if (i >  4 && nidigits <  800 && poly_degree > 11)  break;
+        if (i >  8 && nidigits <  700 && poly_degree >  9)  break;
+        if (i > 16 && nidigits <  600 && poly_degree >  8)  break;
       }
       mpz_set_si(mD, D);
       /* (D/N) must be 1, and we have to have a u,v solution */
@@ -668,20 +786,19 @@ static int ecpp_down(int i, mpz_t Ni, int facstage, UV* dlist, mpz_t* sfacs, int
         continue;
 
       if (verbose > 1)
-        { printf(" %ld", D); fflush(stdout); }
+        { printf(" %d", D); fflush(stdout); }
 
       choose_m(mlist, D, u, v, Ni, t, t2);
       for (k = 0; k < 6; k++) {
-        facresult = check_for_factor2(q, mlist[k], minfactor, t, stage, sfacs, nsfacs);
+        facresult = check_for_factor2(q, mlist[k], minfactor, t, stage, sfacs, nsfacs, poly_degree);
         /* -1 = couldn't find, 0 = no big factors, 1 = found */
         if (facresult <= 0)
           continue;
         mpz_set(m, mlist[k]);
         if (verbose)
-          { printf(" %ld (%s %d)\n", D, (poly_type == 1) ? "Hilbert" : "Weber", poly_degree); fflush(stdout); }
+          { printf(" %d (%s %d)\n", D, (poly_type == 1) ? "Hilbert" : "Weber", poly_degree); fflush(stdout); }
         /* Great, now go down. */
-        /* TODO: This is going to be wasteful */
-        downresult = ecpp_down(i+1, q, stage, dlist, sfacs, nsfacs, prooftextptr);
+        downresult = ecpp_down(i+1, q, next_stage, dlist, sfacs, nsfacs, prooftextptr);
         if (downresult == 0)     /* composite */
           goto end_down;
         if (downresult == 1) {   /* nothing found at this stage */
@@ -695,15 +812,15 @@ static int ecpp_down(int i, mpz_t Ni, int facstage, UV* dlist, mpz_t* sfacs, int
 
         /* Awesome, we found the q chain and are in STAGE 2 */
         if (verbose)
-          { printf("%*sN[%d] (%d dig) %ld (%s %d)", i, "", i, nidigits, D, (poly_type == 1) ? "Hilbert" : "Weber", poly_degree); fflush(stdout); }
+          { printf("%*sN[%d] (%d dig) %d (%s %d)", i, "", i, nidigits, D, (poly_type == 1) ? "Hilbert" : "Weber", poly_degree); fflush(stdout); }
 
         curveresult = find_curve(a, b, P.x, P.y, D, m, q, Ni);
         if (verbose) { printf("  %d\n", curveresult); fflush(stdout); }
         if (curveresult == 1) {
           /* Oh no!  We can't find a point on the curve.  Something is right
            * messed up, and we've wasted a lot of time.  Sigh. */
-          dlist[dnum] = 1; /* skip this D value from now on */
-          if (verbose) gmp_printf("\n  Invalidated D = %ld with N = %Zd\n", D, Ni);
+          dlist[dnum] = -2; /* skip this D value from now on */
+          if (verbose) gmp_printf("\n  Invalidated D = %d with N = %Zd\n", D, Ni);
           downresult = 0;
           continue;
         }
@@ -719,27 +836,64 @@ static int ecpp_down(int i, mpz_t Ni, int facstage, UV* dlist, mpz_t* sfacs, int
 end_down:
 
   if (downresult == 2) {
-    if (verbose > 1) {
-      gmp_printf("N[%lu] = %Zd\n", (unsigned long) i, Ni);
-      gmp_printf("a = %Zd\n", a);
-      gmp_printf("b = %Zd\n", b);
-      gmp_printf("m = %Zd\n", m);
-      gmp_printf("q = %Zd\n", q);
-      gmp_printf("P = (%Zd, %Zd)\n", P.x, P.y);
-      fflush(stdout);
+    if (0 && verbose > 1) {
+      if (D == 1) {
+        gmp_printf("\n");
+        gmp_printf("Type BLS3\n");
+        gmp_printf("N  %Zd\n", Ni);
+        gmp_printf("Q  %Zd\n", q);
+        gmp_printf("A  %lu\n", (unsigned long) nm1a);
+        gmp_printf("\n");
+        fflush(stdout);
+      } else if (D == -1) {
+        gmp_printf("\n");
+        gmp_printf("Type BLS15\n");
+        gmp_printf("N  %Zd\n", Ni);
+        gmp_printf("Q  %Zd\n", q);
+        gmp_printf("LP %ld\n", (signed long) np1lp);
+        gmp_printf("LQ %ld\n", (signed long) np1lq);
+        gmp_printf("\n");
+        fflush(stdout);
+      } else {
+        gmp_printf("\n");
+        gmp_printf("Type ECPP\n");
+        gmp_printf("N  %Zd\n", Ni);
+        gmp_printf("A  %Zd\n", a);
+        gmp_printf("B  %Zd\n", b);
+        gmp_printf("M  %Zd\n", m);
+        gmp_printf("Q  %Zd\n", q);
+        gmp_printf("X  %Zd\n", P.x);
+        gmp_printf("Y  %Zd\n", P.y);
+        gmp_printf("\n");
+        fflush(stdout);
+      }
     }
     /* Prepend our proof to anything that exists. */
     if (prooftextptr != 0) {
       char *proofstr, *proofptr;
-      int myprooflen = mpz_sizeinbase(Ni, 10) * 7 + 20;
       int curprooflen = (*prooftextptr == 0) ? 0 : strlen(*prooftextptr);
 
-      New(0, proofstr, myprooflen + curprooflen + 1, char);
-      proofptr = proofstr;
-      proofptr += gmp_sprintf(proofptr, "%Zd : ECPP : ", Ni);
-      proofptr += gmp_sprintf(proofptr, "%Zd %Zd %Zd %Zd (%Zd:%Zd)\n",
-                                        a, b, m, q, P.x, P.y);
+      if (D == 1) {
+        int myprooflen = 20 + 2*(4 + mpz_sizeinbase(Ni, 10)) + 1*21;
+        New(0, proofstr, myprooflen + curprooflen + 1, char);
+        proofptr = proofstr;
+        proofptr += gmp_sprintf(proofptr, "Type BLS3\nN  %Zd\nQ  %Zd\nA  %"UVuf"\n", Ni, q, nm1a);
+      } else if (D == -1) {
+        int myprooflen = 20 + 2*(4 + mpz_sizeinbase(Ni, 10)) + 2*21;
+        New(0, proofstr, myprooflen + curprooflen + 1, char);
+        proofptr = proofstr;
+        proofptr += gmp_sprintf(proofptr, "Type BLS15\nN  %Zd\nQ  %Zd\nLP %"IVdf"\nLQ %"IVdf"\n", Ni, q, np1lp, np1lq);
+      } else {
+        int myprooflen = 20 + 7*(4 + mpz_sizeinbase(Ni, 10)) + 0;
+        New(0, proofstr, myprooflen + curprooflen + 1, char);
+        proofptr = proofstr;
+        mpz_sub_ui(t, Ni, 1);
+        if (mpz_cmp(a, t) == 0)  mpz_set_si(a, -1);
+        if (mpz_cmp(b, t) == 0)  mpz_set_si(b, -1);
+        proofptr += gmp_sprintf(proofptr, "Type ECPP\nN  %Zd\nA  %Zd\nB  %Zd\nM  %Zd\nQ  %Zd\nX  %Zd\nY  %Zd\n", Ni, a, b, m, q, P.x, P.y);
+      }
       if (*prooftextptr) {
+        proofptr += gmp_sprintf(proofptr, "\n");
         strcat(proofptr, *prooftextptr);
         Safefree(*prooftextptr);
       }
@@ -750,7 +904,7 @@ end_down:
   mpz_clear(a);  mpz_clear(b);
   mpz_clear(u);  mpz_clear(v);
   mpz_clear(m);  mpz_clear(q);
-  mpz_clear(mD); mpz_clear(minfactor);
+  mpz_clear(mD); mpz_clear(minfactor);  mpz_clear(sqrtn);
   mpz_clear(t);  mpz_clear(t2);
   mpz_clear(P.x);mpz_clear(P.y);
   for (k = 0; k < 6; k++)
@@ -763,7 +917,7 @@ end_down:
 /* returns 2 if N is proven prime, 1 if probably prime, 0 if composite */
 int _GMP_ecpp(mpz_t N, char** prooftextptr)
 {
-  UV* dlist;
+  IV* dlist;
   mpz_t* sfacs;
   int i, fstage, result, nsfacs;
   int verbose = get_verbose_level();
@@ -772,11 +926,13 @@ int _GMP_ecpp(mpz_t N, char** prooftextptr)
   if (mpz_gcd_ui(NULL, N, 223092870UL) != 1)
     return _GMP_is_prob_prime(N);
 
+  init_ecpp_gcds();
+
   if (prooftextptr)
     *prooftextptr = 0;
 
   New(0, sfacs, MAX_SFACS, mpz_t);
-  dlist = poly_class_degrees();
+  dlist = poly_class_degrees(1);
   nsfacs = 0;
   result = 1;
   for (fstage = 1; fstage < 20; fstage++) {
@@ -813,7 +969,7 @@ int _GMP_ecpp_fps(mpz_t N, char** prooftextptr)
 {
   mpz_t Ni, a, b, u, v, m, q, mD, t, t2, minfactor;
   mpz_t mlist[6];
-  UV* dlist;
+  IV* dlist;
   struct ec_affine_point P;
   UV i, dnum;
   long D, k, numQs, numDs, stage;
@@ -828,6 +984,8 @@ int _GMP_ecpp_fps(mpz_t N, char** prooftextptr)
   if (mpz_gcd_ui(NULL, N, 223092870UL) != 1)
     return _GMP_is_prob_prime(N);
 
+  init_ecpp_gcds();
+
   if (prooftextptr) {
     proofstr_size = 4096;
     New(0, proofstr, proofstr_size, char);
@@ -845,7 +1003,7 @@ int _GMP_ecpp_fps(mpz_t N, char** prooftextptr)
   mpz_init(t); mpz_init(t2);
   mpz_init(minfactor);
   mpz_init(P.x); mpz_init(P.y);
-  dlist = poly_class_degrees();
+  dlist = poly_class_degrees(0);
   for (k = 0; k < 6; k++)
     mpz_init(mlist[k]);
   /* TODO: make this 100 and realloc if needed */
@@ -1091,6 +1249,8 @@ int main(int argc, char **argv)
   int isprime, i, do_printcert;
   int do_nminus1 = 0;
   int do_aks = 0;
+  int do_aprcl = 0;
+  int do_bpsw = 0;
   char* cert = 0;
 
   if (argc < 2) dieusage(argv[0]);
@@ -1112,6 +1272,10 @@ int main(int argc, char **argv)
         do_nminus1 = 1;
       } else if (strcmp(argv[i], "-aks") == 0) {
         do_aks = 1;
+      } else if (strcmp(argv[i], "-aprcl") == 0) {
+        do_aprcl = 1;
+      } else if (strcmp(argv[i], "-bpsw") == 0) {
+        do_bpsw = 1;
       } else if (strcmp(argv[i], "-help") == 0 || strcmp(argv[i], "--help") == 0) {
         dieusage(argv[0]);
       } else {
@@ -1121,20 +1285,30 @@ int main(int argc, char **argv)
       continue;
     }
     mpz_set_str(n, argv[i], 10);
-    gmp_printf("%Zd\n", n);
+    /* gmp_printf("%Zd\n", n); */
 
     isprime = _GMP_is_prob_prime(n);
     /* If isprime = 2 here, that means it's so small it fits in the
      * deterministic M-R or BPSW range. */
     if (isprime == 2) {
-      Newz(0, cert, 4 + mpz_sizeinbase(n, 10), char);
-      gmp_sprintf(cert, "[%Zd]\n", n);
+      Newz(0, cert, 20 + mpz_sizeinbase(n, 10), char);
+      gmp_sprintf(cert, "Type Small\nN  %Zd\n", n);
     } else if (isprime == 1) {
-      if (do_nminus1) {
+      if (do_bpsw) {
+        /* Done */
+      } else if (do_nminus1) {
         isprime = _GMP_primality_bls_nm1(n, 100, &cert);
       } else if (do_aks) {
-        isprime = _GMP_is_aks_prime(n);
+        isprime = 2 * _GMP_is_aks_prime(n);
         do_printcert = 0;
+      } else if (do_aprcl) {
+#ifdef USE_APRCL
+        /* int i; for (i = 0; i < 10000; i++) */
+        isprime = mpz_aprtcle(n, get_verbose_level());
+        do_printcert = 0;
+#else
+        croak("Compiled without USE_APRCL.  Sorry.");
+#endif
       } else {
         /* Quick n-1 test */
         isprime = _GMP_primality_bls_nm1(n, 1, &cert);
@@ -1143,18 +1317,29 @@ int main(int argc, char **argv)
       }
     }
 
+    /* printf("(%d digit) ", (int)mpz_sizeinbase(n, 10)); */
     if (isprime == 0) {
       printf("COMPOSITE\n");
     } else if (isprime == 1) {
       /* We really shouldn't ever see this. */
       printf("PROBABLY PRIME\n");
     } else {
-      printf("PRIME\n");
-      if (do_printcert)
+      if (do_printcert) {
+        gmp_printf("[MPU - Primality Certificate]\n");
+        gmp_printf("Version 1.0\n");
+        gmp_printf("\n");
+        gmp_printf("Proof for:\n");
+        gmp_printf("N %Zd\n", n);
+        gmp_printf("\n");
         printf("%s", cert);
+      } else {
+        printf("PRIME\n");
+      }
     }
-    if (cert != 0)
+    if (cert != 0) {
       Safefree(cert);
+      cert = 0;
+    }
   }
   mpz_clear(n);
   _GMP_destroy();
